import time
import socket
import random
import logging
import warnings
from gzip import GzipFile
from itertools import count
import requests
from requests.compat import StringIO
from collections import deque
from Queue import Queue
from threading import Thread, Event
from .utils import xauth, iterqueue
from .serialization import jsonencode

logger = logging.getLogger('hubstorage.batchuploader')


class BatchUploader(object):

    # Wait time between all batches status checks
    worker_loop_delay = 1.0
    # Max number of retries attempts before giving up,
    # Using 10 minutes max interval it accounts for 30hs total wait time
    worker_max_retries = 200
    # Minimum and maximum wait time between retries in seconds,
    # it can be 50% greater due to randomization
    worker_min_interval = 30
    worker_max_interval = 600

    def __init__(self, client):
        self.client = client
        self.closed = False
        self._wait_event = Event()
        self._writers = deque()
        self._thread = Thread(target=self._worker)
        self._thread.daemon = True
        self._thread.start()

    def create_writer(self, url, start=0, auth=None, size=1000, interval=15,
                      qsize=None, content_encoding='identity',
                      maxitemsize=1024 ** 2, callback=None):
        assert not self.closed, 'Can not create new writers when closed'
        auth = xauth(auth) or self.client.auth
        w = _BatchWriter(url=url,
                         auth=auth,
                         size=size,
                         start=start,
                         interval=interval,
                         qsize=qsize,
                         maxitemsize=maxitemsize,
                         content_encoding=content_encoding,
                         uploader=self,
                         callback=callback)
        self._writers.append(w)
        return w

    def close(self, timeout=None):
        self.closed = True
        self.interrupt()
        self._thread.join(timeout)

    def interrupt(self):
        self._wait_event.set()

    def __del__(self):
        if not self.closed:
            warnings.warn("%r not closed properly, some items may have been "
                          "lost!: %r" % (self.__class__.__name__, self._writers))

    def _interruptable_sleep(self):
        self._wait_event.wait(self.worker_loop_delay)
        self._wait_event.clear()

    def _worker(self):
        ctr = count()
        while True:
            if not self._writers:
                # Stop thread if closed and idle, but if open wait for writers
                if self.closed:
                    break
                self._interruptable_sleep()
                continue

            # Delay once all writers are processed
            if (ctr.next() % len(self._writers) == 0) and not self.closed:
                self._interruptable_sleep()

            # Get next writer to process
            w = self._writers.popleft()

            # Close open writers if uploader is closed
            if self.closed and not w.closed:
                w.close(block=False)

            # Checkpoint writer if eligible
            now = time.time()
            if w.itemsq.qsize() >= w.size or w.closed or w.flushme \
                    or w.checkpoint < now - w.interval:
                self._checkpoint(w)
                w.checkpoint = now

            # Re-queue pending or open writers
            if not (w.closed and w.itemsq.empty()):
                self._writers.append(w)

    def _checkpoint(self, w):
        q = w.itemsq
        qiter = iterqueue(q, w.size)
        data = self._content_encode(qiter, w)
        if qiter.count > 0:
            response = self._tryupload({
                'url': w.url,
                'offset': w.offset,
                'data': data,
                'auth': w.auth,
                'content-encoding': w.content_encoding,
            })
            w.offset += qiter.count
            for _ in xrange(qiter.count):
                q.task_done()
            if w.callback is not None:
                try:
                    w.callback(response)
                except Exception, e:
                    logger.exception("Callback for %s failed", w.url)

    def _content_encode(self, qiter, w):
        ce = w.content_encoding
        if ce == 'identity':
            return _encode_identity(qiter)
        elif ce == 'gzip':
            return _encode_gzip(qiter)
        else:
            raise ValueError('Writer using unknown content encoding: %s' % ce)

    def _tryupload(self, batch):
        """Retry uploads in case of server failures

        Use polinomial backoff with 10 minutes maximum interval that accounts
        for ~30 hours of total retry time.

        >>> sum(min(x**2, 600) for x in xrange(200)) / 3600
        30
        """
        url = batch['url']
        offset = batch['offset']
        for retryn in xrange(self.worker_max_retries):
            emsg = ''
            try:
                r = self._upload(batch)
                r.raise_for_status()
                if not (200 <= r.status_code < 300):
                    logger.warning('Discarding write to url=%s offset=%s: '
                                   '[HTTP error %s] %s\n%s', url, offset,
                                   r.status_code, r.reason, r.text.rstrip())
                return r
            except (socket.error, requests.RequestException) as e:
                if isinstance(e, requests.HTTPError):
                    emsg = "[HTTP error {0}] {1}".format(e.response.status_code,
                                                         e.response.text.rstrip())
                else:
                    emsg = str(e)
                logger.info("Retrying url=%s offset=%s: %s", url, offset, emsg)
            except Exception:
                logger.exception('Non retryable failure on url=%s offset=%s',
                                 url, offset)
                break

            backoff = min(max(retryn ** 2, self.worker_min_interval),
                          self.worker_max_interval)
            time.sleep(backoff * (0.5 + random.random()))

    def _upload(self, batch):
        params = {'start': batch['offset']}
        headers = {'content-encoding': batch['content-encoding']}
        return self.client.session.request(
            method='POST',
            url=batch['url'],
            data=batch['data'],
            auth=batch['auth'],
            timeout=self.client.connection_timeout,
            params=params,
            headers=headers,
        )

class ValueTooLarge(ValueError):
    """Raised when a serialized item is greater than 1MB"""


class _BatchWriter(object):
    #: Truncate overly big items to that many bytes for the error message.
    ERRMSG_DATA_TRUNCATION_LEN = 1024

    def __init__(self, url, start, auth, size, interval, qsize,
                 maxitemsize, content_encoding, uploader, callback=None):
        self.url = url
        self.offset = start
        self._nextid = count(start)
        self.auth = auth
        self.size = size
        self.interval = interval
        self.maxitemsize = maxitemsize
        self.content_encoding = content_encoding
        self.checkpoint = time.time()
        self.itemsq = Queue(size * 2 if qsize is None else qsize)
        self.closed = False
        self.flushme = False
        self.uploader = uploader
        self.callback = callback

    def write(self, item):
        assert not self.closed, 'attempting writes to a closed writer'
        data = jsonencode(item)
        if len(data) > self.maxitemsize:
            truncated_data = data[:self.ERRMSG_DATA_TRUNCATION_LEN] + "..."
            raise ValueTooLarge(
                'Value exceeds max encoded size of {} bytes: {!r}'
                .format(self.maxitemsize, truncated_data))

        self.itemsq.put(data)
        if self.itemsq.full():
            self.uploader.interrupt()
        return self._nextid.next()

    def flush(self):
        self.flushme = True
        self._waitforq()
        self.flushme = False

    def close(self, block=True):
        self.closed = True
        if block:
            self._waitforq()

    def _waitforq(self):
        self.uploader.interrupt()
        self.itemsq.join()

    def __str__(self):
        return self.url


def _encode_identity(iter):
    data = StringIO()
    for item in iter:
        data.write(item)
        data.write('\n')
    return data.getvalue()


def _encode_gzip(iter):
    data = StringIO()
    with GzipFile(fileobj=data, mode='w') as gzo:
        for item in iter:
            gzo.write(item)
            gzo.write('\n')
    return data.getvalue()



import random
from contextlib import closing
from hstestcase import HSTestCase
from hubstorage import HubstorageClient
from hubstorage.utils import millitime


class SystemTest(HSTestCase):

    MAGICN = 1211

    def setUp(self):
        super(HSTestCase, self).setUp()
        self.endpoint = self.hsclient.endpoint
        # Panel - no client auth, only project auth using user auth token
        self.panelclient = HubstorageClient(endpoint=self.endpoint)
        self.panelproject = self.panelclient.get_project(self.projectid, auth=self.auth)

    def tearDown(self):
        super(HSTestCase, self).tearDown()
        self.panelclient.close()

    def test_succeed_with_close_reason(self):
        self._do_test_success('all-good', 'all-good')

    def test_succeed_without_close_reason(self):
        self._do_test_success(None, 'no_reason')

    def test_scraper_failure(self):
        job = self._do_test_job(IOError('no more resources, ha!'), 'failed')
        # MAGICN per log level messages plus one of last failure
        stats = job.logs.stats()
        self.assertTrue(stats)
        self.assertEqual(stats['totals']['input_values'], self.MAGICN * 4 + 1)

    def _do_test_success(self, job_close_reason, expected_close_reason):
        job = self._do_test_job(job_close_reason, expected_close_reason)
        self.assertEqual(job.items.stats()['totals']['input_values'], self.MAGICN)
        self.assertEqual(job.logs.stats()['totals']['input_values'], self.MAGICN * 4)
        self.assertEqual(job.requests.stats()['totals']['input_values'], self.MAGICN)

    def _do_test_job(self, job_close_reason, expected_close_reason):
        p = self.panelproject
        pushed = p.jobq.push(self.spidername)
        # check pending state
        job = p.get_job(pushed['key'])
        self.assertEqual(job.metadata.get('state'), 'pending')
        # consume msg from runner
        self._run_runner(pushed, close_reason=job_close_reason)
        # query again from panel
        job = p.get_job(pushed['key'])
        self.assertEqual(job.metadata.get('state'), 'finished')
        self.assertEqual(job.metadata.get('close_reason'), expected_close_reason)
        return job

    def _run_runner(self, pushed, close_reason):
        client = HubstorageClient(endpoint=self.endpoint, auth=self.auth)
        with closing(client) as runnerclient:
            job = runnerclient.start_job(self.projectid)
            self.assertFalse(job.metadata.get('stop_requested'))
            job.metadata.update(host='localhost', slot=1)
            self.assertEqual(job.metadata.get('state'), 'running')
            # run scraper
            try:
                self._run_scraper(job.key, job.jobauth, close_reason=close_reason)
            except Exception as exc:
                job.failed(message=str(exc))
                # logging from runner must append and never remove messages logged
                # by scraper
                self.assertTrue(job.logs.batch_append)
            else:
                job.finished()

    def _run_scraper(self, jobkey, jobauth, close_reason=None):
        httpmethods = 'GET PUT POST DELETE HEAD OPTIONS TRACE CONNECT'.split()
        # Scraper - uses job level auth, no global or project auth available
        client = HubstorageClient(endpoint=self.endpoint)
        with closing(client) as scraperclient:
            job = scraperclient.get_job(jobkey, auth=jobauth)
            for idx in xrange(self.MAGICN):
                iid = job.items.write({'uuid': idx})
                job.logs.debug('log debug %s' % idx, idx=idx)
                job.logs.info('log info %s' % idx, idx=idx)
                job.logs.warn('log warn %s' % idx, idx=idx)
                job.logs.error('log error %s' % idx, idx=idx)
                sid = job.samples.write([idx, idx, idx])
                rid = job.requests.add(
                    url='http://test.com/%d' % idx,
                    status=random.randint(100, 1000),
                    method=random.choice(httpmethods),
                    rs=random.randint(0, 100000),
                    duration=random.randint(0, 1000),
                    parent=random.randrange(0, idx + 1) if idx > 10 else None,
                    ts=millitime() + random.randint(100, 100000),
                )
                self.assertEqual(iid, idx)
                self.assertEqual(sid, idx)
                self.assertEqual(rid, idx)

            if isinstance(close_reason, Exception):
                raise close_reason

            if close_reason:
                job.metadata['close_reason'] = close_reason

            job.metadata.save()



from contextlib import contextmanager
from requests import Timeout


@contextmanager
def failing_downloader(downloader, N=5, exc=Timeout, msg='test error'):
    # reduce wait times and simulate network timeout
    orig_iter = downloader._iter_lines
    orig_wait = downloader.RETRY_INTERVAL
    # generate a Timeout every N requests
    downloader._iter_lines = wrap_seq_to_fail(
        downloader._iter_lines, exc, N, msg)
    downloader.RETRY_INTERVAL = 0
    try:
        yield downloader
    finally:
        downloader._iter_lines = orig_iter
        downloader.RETRY_INTERVAL = orig_wait


def wrap_seq_to_fail(func, exception, N, *exc_args):
    """Wrap a function that returns a sequence with failafter"""
    def _wrapper(*args, **kwargs):
        seq = func(*args, **kwargs)
        return failafter(seq, exception, N, *exc_args)
    return _wrapper


def failafter(seq, exception, N, *exc_args):
    """iterate over seq, and raise exception after N items"""
    for i, item in enumerate(seq):
        if i == N:
            raise exception(*exc_args)
        yield item



import logging
from .resourcetype import (ItemsResourceType, DownloadableResource,
    MappingResourceType)
from .utils import millitime, urlpathjoin
from .jobq import JobQ


class Job(object):

    def __init__(self, client, key, auth=None, jobauth=None, metadata=None):
        self.key = urlpathjoin(key)
        assert len(self.key.split('/')) == 3, \
            'Jobkey must be projectid/spiderid/jobid: %s' % self.key
        self.jobauth = jobauth
        self.auth = self.jobauth or auth
        self.metadata = JobMeta(client, self.key, self.auth, cached=metadata)
        self.items = Items(client, self.key, self.auth)
        self.logs = Logs(client, self.key, self.auth)
        self.samples = Samples(client, self.key, self.auth)
        self.requests = Requests(client, self.key, self.auth)
        self.jobq = JobQ(client, self.key.split('/')[0], auth)

    def close_writers(self):
        wl = [self.items, self.logs, self.samples, self.requests]
        # close all resources that use background writers
        for w in wl:
            w.close(block=False)
        # now wait for all writers to close together
        for w in wl:
            w.close(block=True)

    def update_metadata(self, *args, **kwargs):
        self.metadata.update(*args, **kwargs)
        self.metadata.save()
        self.metadata.expire()

    def request_cancel(self):
        self.jobq.request_cancel(self)

    def finished(self, close_reason=None):
        self.metadata.expire()
        close_reason = close_reason or \
            self.metadata.liveget('close_reason') or 'no_reason'
        self.update_metadata(close_reason=close_reason)
        self.close_writers()
        self.jobq.finish(self)

    def failed(self, reason='failed', message=None):
        if message:
            self.logs.error(message, appendmode=True)
        self.finished(reason)

    def purged(self):
        self.jobq.delete(self)
        self.metadata.expire()


class JobMeta(MappingResourceType):

    resource_type = 'jobs'
    ignore_fields = set(('auth', '_key', 'state'))

    def authtoken(self):
        return self.liveget('auth')


class Logs(ItemsResourceType, DownloadableResource):

    resource_type = 'logs'
    batch_content_encoding = 'gzip'

    def __init__(self, client, key, auth=None, appendmode=False):
        ItemsResourceType.__init__(self, client, key, auth)
        self.batch_append = appendmode

    def batch_write_start(self):
        if self.batch_append:
            return self.stats().get('totals', {}).get('input_values', 0)
        return 0

    def log(self, message, level=logging.INFO, ts=None, **other):
        other.update(message=message, level=level, time=ts or millitime())
        # legacy support for an appendmode argument. This should be set at
        # object initialization time.
        if self._writer is None and other.get('appendmode'):
            self.batch_append = True
        self.write(other)

    def debug(self, message, **other):
        self.log(message, level=logging.DEBUG, **other)

    def info(self, message, **other):
        self.log(message, level=logging.INFO, **other)

    def warn(self, message, **other):
        self.log(message, level=logging.WARNING, **other)
    warning = warn

    def error(self, message, **other):
        self.log(message, level=logging.ERROR, **other)


class Items(ItemsResourceType, DownloadableResource):

    resource_type = 'items'
    batch_content_encoding = 'gzip'


class Samples(ItemsResourceType):

    resource_type = 'samples'

    def stats(self):
        raise NotImplementedError('Resource does not expose stats')


class Requests(ItemsResourceType, DownloadableResource):

    resource_type = 'requests'
    batch_content_encoding = 'gzip'

    def add(self, url, status, method, rs, parent, duration, ts, fp=None):
        return self.write({
            'url': url,
            'status': int(status),
            'method': method,
            'rs': int(rs),
            'duration': int(duration),
            'parent': parent,
            'time': int(ts),
            'fp': fp,
        })



import json
from requests.exceptions import HTTPError
from .resourcetype import ResourceType


class DuplicateJobError(Exception):
    """Raised when a job with same unique is pushed"""


class JobQ(ResourceType):

    resource_type = 'jobq'

    PRIO_LOWEST = 0
    PRIO_LOW = 1
    PRIO_NORMAL = 2
    PRIO_HIGH = 3
    PRIO_HIGHEST = 4

    def push(self, spider, **jobparams):
        jobparams['spider'] = spider
        try:
            for o in self.apipost('push', jl=jobparams):
                if 'error' in o:
                    if 'Active job' in o['error']:
                        raise DuplicateJobError(o['error'])
                    raise HTTPError(o['error'])
                return o
        except HTTPError as exc:
            if exc.response and exc.response.status_code == 409:
                raise DuplicateJobError()
            raise

    def summary(self, _queuename=None, spiderid=None, count=None, start=None, jobmeta=None):
        params = {}
        if count is not None:
            params['count'] = count
        if start is not None:
            params['start'] = start
        if jobmeta is not None:
            params['jobmeta'] = jobmeta

        r = list(self.apiget((spiderid, 'summary', _queuename), params=params))
        return (r and r[0] or None) if _queuename else r

    def list(self, spider=None, count=None, stop=None, state=None,
             has_tag=None, lacks_tag=None, startts=None, endts=None,
             **params):
        if 'filter' in params:
            return self._legacy_list_with_filter(params)

        if state is not None:
            params['state'] = state
        if spider is not None:
            params['spider'] = spider
        if count is not None:
            params['count'] = count
        if stop is not None:
            params['stop'] = stop
        if startts is not None:
            params['startts'] = startts
        if endts is not None:
            params['endts'] = endts
        if has_tag is not None:
            params['has_tag'] = has_tag
        if lacks_tag is not None:
            params['lacks_tag'] = lacks_tag
        return self.apiget(('list',), params=params)

    def _legacy_list_with_filter(self, params):
        only_finished_outcome = False
        for row in params['filter']:
            field, matchdecider, value = json.loads(row)
            if field == 'tags' and matchdecider == 'haselement':
                params['has_tag'] = value
            if field == 'tags' and matchdecider == 'hasnotelement':
                params['lacks_tag'] = value
            if field == 'state' and matchdecider == '=':
                params['state'] = value
            if field == 'spider' and matchdecider == '=':
                params['spider'] = value
            if field == 'close_reason' and value == ['finished']:
                only_finished_outcome = True
                params.setdefault('jobmeta', []).append('close_reason')

        jobs = self.apiget(('list',), params=params)
        if only_finished_outcome:
            return (x for x in jobs if x.get('close_reason') == 'finished')
        return jobs

    def start(self, job=None, **start_params):
        """Start a new job

        If a job is passed, it is changed to the started state and metadata
        updated with the start_params. Otherwise the next job is pulled from
        hubstorage, using the start_params which will be saved as metadata.

        If a 'botgroup' parameter is present in start_params, only jobs from
        that botgroup will be started.

        It may take up to a second for a previously added job to be available.
        """
        if job:
            return self._set_state(job, 'running', **start_params)
        for o in self.apipost('startjob', jl=start_params):
            return o

    def request_cancel(self, job):
        """Cancel a running job"""
        self.apipost("%s/cancel" % job.key[job.key.index('/') + 1:])

    def finish(self, job):
        return self._set_state(job, 'finished')

    def delete(self, job):
        return self._set_state(job, 'deleted')

    def _jobkeys(self, job):
        if isinstance(job, list):
            for x in job:
                for k in self._jobkeys(x):
                    yield k
        elif isinstance(job, dict):
            yield job['key']
        elif hasattr(job, 'key'):
            yield job.key
        else:
            yield job

    def _set_state(self, job, state, **extra_args):
        data = [dict(extra_args, key=k, state=state) for k in self._jobkeys(job)]
        return self.apipost('update', jl=data)



"""
Test Collections
"""
import random
from contextlib import closing
from hstestcase import HSTestCase
from testutil import failing_downloader


def _mkitem():
    return dict(field1='value1', field2=['value2a', 'value2b'], field3=3, field4={'v4k': 'v4v'})


class CollectionsTest(HSTestCase):

    test_collection_name = "test_collection_%s" % random.randint(1, 1000000)

    def post_get_delete_test(self):
        test_item = _mkitem()
        item_to_send = dict(test_item)
        item_to_send['_key'] = test_key = 'insert_test_key'

        test_collections = [
            self.project.collections.new_store(self.test_collection_name),
            self.project.collections.new_cached_store(self.test_collection_name),
            self.project.collections.new_versioned_store(self.test_collection_name),
            self.project.collections.new_versioned_cached_store(self.test_collection_name),
        ]

        test_collections.extend(
            self.project.collections.new_collection(t, self.test_collection_name + 'b')
            for t in ('s', 'vs', 'cs', 'vcs'))

        for col in test_collections:
            col.set(item_to_send)
            returned_item = col.get(test_key)
            self.assertEqual(test_item, returned_item)
            col.delete(test_key)
            self.assertRaises(KeyError, col.get, test_key)

    def post_scan_test(self):
        col = self.project.collections.new_store(self.test_collection_name)

        # populate with 20 items
        test_item = _mkitem()
        last_key = None
        with closing(col.create_writer()) as writer:
            for i in xrange(20):
                test_item['_key'] = last_key = "post_scan_test%d" % i
                test_item['counter'] = i
                writer.write(test_item)

        # check last value is as expected
        returned_item = col.get(last_key)
        del test_item['_key']
        self.assertEqual(test_item, returned_item)

        # get all values starting with 1
        result = list(col.get(prefix='post_scan_test1'))
        # 1 & 10-19 = 11 items
        self.assertEqual(len(result), 11)

        # combining with normal filters
        result = list(col.get(filter='["counter", ">", [5]]', prefix='post_scan_test1'))
        # 10-19
        self.assertEqual(len(result), 10)

        # bulk delete
        col.delete('post_scan_test%d' % i for i in xrange(20))

        # test items removed (check first and last)
        self.assertRaises(KeyError, col.get, 'post_scan_test0')
        self.assertRaises(KeyError, col.get, last_key)

    def test_errors(self):
        col = self.project.collections.new_store(self.test_collection_name)
        self.assertRaises(KeyError, col.get, 'does_not_exist')
        self.assertRaises(ValueError, col.set, {'foo': 42})
        self.assertRaises(ValueError, col.set, {'_key': []})
        self.assertRaises(ValueError, col.set,
            {'_key': 'large_test', 'value': 'x' * 1024 ** 2})

    def test_data_download(self):
        col = self.project.collections.new_store(self.test_collection_name)
        items = []
        with closing(col.create_writer()) as writer:
            for i in xrange(20):
                test_item = _mkitem()
                test_item['_key'] = "test_data_download%d" % i
                test_item['counter'] = i
                writer.write(test_item)
                items.append(test_item)

        # check parameters are passed correctly
        downloaded = list(col.iter_values(prefix='test_data_download1'))
        self.assertEqual(len(downloaded), 11)

        # simulate network timeouts and download data
        with failing_downloader(self.project.collections):
            downloaded = list(col.iter_values(start='test_data_download1'))
            self.assertEqual(len(downloaded), 19)



import warnings

from .job import Job
from .jobq import JobQ
from .activity import Activity
from .collectionsrt import Collections
from .frontier import Frontier
from .resourcetype import ResourceType, MappingResourceType
from .utils import urlpathjoin, xauth


class Project(object):

    def __init__(self, client, projectid, auth=None):
        self.client = client
        self.projectid = urlpathjoin(projectid)
        assert len(self.projectid.split('/')) == 1, \
                'projectkey must be just one id: %s' % projectid
        self.auth = xauth(auth) or client.auth
        self.jobs = Jobs(client, self.projectid, auth=auth)
        self.items = Items(client, self.projectid, auth=auth)
        self.logs = Logs(client, self.projectid, auth=auth)
        self.samples = Samples(client, self.projectid, auth=auth)
        self.jobq = JobQ(client, self.projectid, auth=auth)
        self.activity = Activity(client, self.projectid, auth=auth)
        self.collections = Collections(client, self.projectid, auth=auth)
        self.frontier = Frontier(client, self.projectid, auth=auth)
        self.ids = Ids(client, self.projectid, auth=auth)
        self.settings = Settings(client, self.projectid, auth=auth)
        self.reports = Reports(client, self.projectid, auth=auth)
        self.spiders = Spiders(client, self.projectid, auth=auth)

    def get_job(self, _key, *args, **kwargs):
        key = urlpathjoin(_key)
        parts = key.split('/')
        if len(parts) == 2:
            key = (self.projectid, key)
        elif len(parts) == 3 and parts[0] == self.projectid:
            pass
        else:
            raise ValueError('Invalid jobkey %s for project %s'
                             % (key, self.projectid))

        kwargs.setdefault('auth', self.auth)
        return self.client.get_job(key, *args, **kwargs)

    def get_jobs(self, **kwargs):
        warnings.warn('Method `project.get_jobs()` is deprecated, '
                      'use `project.jobq.list()` instead', Warning)
        for metadata in self.jobq.list(**kwargs):
            key = metadata.pop('key')
            yield self.get_job(key, metadata=metadata)

    def push_job(self, spidername, **jobparams):
        data = self.jobq.push(spidername, **jobparams)
        key = data['key']
        return Job(self.client, key, auth=self.auth)

    def start_job(self, **startparams):
        return self.client.start_job(projectid=self.projectid, auth=self.auth,
            **startparams)

    def jobsummary(self, **params):
        uri = ('projects', self.projectid, 'jobsummary')
        return next(self.client.root.apiget(uri, auth=self.auth, params=params))


class Jobs(ResourceType):

    resource_type = 'jobs'

    def list(self, _key=None, **params):
        return self.apiget(_key, params=params)

class Items(ResourceType):

    resource_type = 'items'

    def list(self, _key=None, **params):
        return self.apiget(_key, params=params)


class Logs(ResourceType):

    resource_type = 'logs'

    def list(self, _key=None, **params):
        return self.apiget(_key, params=params)


class Samples(ResourceType):

    resource_type = 'samples'

    def list(self, _key=None, **params):
        return self.apiget(_key, params=params)


class Ids(ResourceType):

    resource_type = 'ids'

    def spider(self, spidername, **params):
        r = self.apiget(('spider', spidername), params=params)
        return r.next()


class Settings(MappingResourceType):

    resource_type = 'projects'
    key_suffix = 'settings'


class Reports(ResourceType):

    resource_type = 'projects'
    key_suffix = 'reports'


class Spiders(ResourceType):

    resource_type = 'spiders'

    def lastjobsummary(self, spiderid=None, **params):
        return self.apiget((spiderid, 'lastjobsummary'), params=params)



import logging, time, json, socket
from collections import MutableMapping
import requests.exceptions as rexc
from .utils import urlpathjoin, xauth
from .serialization import jlencode, jldecode

logger = logging.getLogger('hubstorage.resourcetype')


class ResourceType(object):

    resource_type = None
    key_suffix = None

    def __init__(self, client, key, auth=None):
        self.client = client
        self.key = urlpathjoin(self.resource_type, key, self.key_suffix)
        self.auth = xauth(auth) or client.auth
        self.url = urlpathjoin(client.endpoint, self.key)

    def _iter_lines(self, _path, **kwargs):
        kwargs['url'] = urlpathjoin(self.url, _path)
        kwargs.setdefault('auth', self.auth)
        if 'jl' in kwargs:
            kwargs['data'] = jlencode(kwargs.pop('jl'))

        r = self.client.request(**kwargs)

        return r.iter_lines()

    def apirequest(self, _path=None, **kwargs):
        return jldecode(self._iter_lines(_path, **kwargs))

    def apipost(self, _path=None, **kwargs):
        return self.apirequest(_path, method='POST', **kwargs)

    def apiget(self, _path=None, **kwargs):
        kwargs.setdefault('is_idempotent', True)
        return self.apirequest(_path, method='GET', **kwargs)

    def apidelete(self, _path=None, **kwargs):
        kwargs.setdefault('is_idempotent', True)
        return self.apirequest(_path, method='DELETE', **kwargs)


class DownloadableResource(ResourceType):
    MAX_RETRIES = 180
    RETRY_INTERVAL = 60

    def _add_resume_param(self, lastline, offset, params):
        """Adds a startafter=LASTKEY parameter if there was
        a lastvalue. It also adds meta=_key to ensure a key is returned
        """
        meta = params.get('meta', [])
        if '_key' not in meta:
            meta = list(meta)
            meta.append('_key')
            params['meta'] = meta
        if lastline is not None:
            lastvalue = json.loads(lastline)
            params['startafter'] = lastvalue['_key']
            if 'start' in params:
                del params['start']

    def iter_values(self, *args, **kwargs):
        """Reliably iterate through all data as python objects

        calls iter_json, decoding the results
        """
        return jldecode(self.iter_json(*args, **kwargs))

    def iter_json(self, _path=None, requests_params=None, **apiparams):
        """Reliably iterate through all data as json strings"""
        requests_params = dict(requests_params or {})
        requests_params.setdefault('method', 'GET')
        requests_params.setdefault('stream', True)
        lastexc = None
        line = None
        offset = 0
        for attempt in xrange(self.MAX_RETRIES):
            self._add_resume_param(line, offset, apiparams)
            try:
                for line in self._iter_lines(_path=_path, params=apiparams,
                        **requests_params):
                    yield line
                    offset += 1
                break
            except (ValueError, socket.error, rexc.RequestException) as exc:
                # catch requests exceptions other than HTTPError
                if isinstance(exc, rexc.HTTPError):
                    raise
                lastexc = exc
                url = urlpathjoin(self.url, _path)
                msg = "Retrying read of %s in %ds: attempt=%d/%d error=%s"
                args = url, self.RETRY_INTERVAL, attempt, self.MAX_RETRIES, exc
                logger.debug(msg, *args)
                time.sleep(self.RETRY_INTERVAL)
        else:
            url = urlpathjoin(self.url, _path)
            logger.error("Failed %d times reading items from %s, params %s, "
                "last error was: %s", self.MAX_RETRIES, url, apiparams, lastexc)


class ItemsResourceType(ResourceType):

    batch_size = 1000
    batch_qsize = None  # defaults to twice batch_size if None
    batch_start = 0
    batch_interval = 15.0
    batch_content_encoding = 'identity'

    # batch writer reference in case of used
    _writer = None

    # TODO override _add_resume_param - can avoid requestomg _key by
    # deriving from project, spider, job and offset

    def batch_write_start(self):
        """Override to set a start parameter when commencing writing"""
        return 0

    @property
    def writer(self):
        if self._writer is None:
            self._writer = self.client.batchuploader.create_writer(
                url=self.url,
                auth=self.auth,
                size=self.batch_size,
                start=self.batch_write_start(),
                interval=self.batch_interval,
                qsize=self.batch_qsize,
                content_encoding=self.batch_content_encoding
            )
        return self._writer

    def flush(self):
        if self._writer is not None:
            self._writer.flush()

    def close(self, block=True):
        if self._writer is not None:
            self._writer.close(block=block)

    def write(self, item):
        return self.writer.write(item)

    def list(self, _key=None, **params):
        return self.apiget(_key, params=params)

    def get(self, _key, **params):
        """Return first matching result"""
        for o in self.list(_key, params=params):
            return o

    def stats(self):
        return self.apiget('stats').next()


class MappingResourceType(ResourceType, MutableMapping):

    _cached = None
    ignore_fields = ()

    def __init__(self, *a, **kw):
        self._cached = kw.pop('cached', None)
        self._deleted = set()
        super(MappingResourceType, self).__init__(*a, **kw)

    def __str__(self):
        return str(self._data)

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, repr(self._data))

    @property
    def _data(self):
        if self._cached is None:
            r = self.apiget()
            try:
                self._cached = r.next()
            except StopIteration:
                self._cached = {}

        return self._cached

    def expire(self):
        self._cached = None

    def save(self):
        for key in self._deleted:
            self.apidelete(key)
        self._deleted.clear()
        if self._cached:
            if not self.ignore_fields:
                self.apipost(jl=self._data, is_idempotent=True)
            else:
                self.apipost(jl=dict((k, v) for k, v in self._data.iteritems()
                                     if k not in self.ignore_fields),
                             is_idempotent=True)

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        self._data[key] = value
        self._deleted.discard(key)

    def __delitem__(self, key):
        del self._data[key]
        self._deleted.add(key)

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def liveget(self, key):
        for o in self.apiget(key):
            return o



from json import dumps, loads
from datetime import datetime

EPOCH = datetime.utcfromtimestamp(0)
ADAYINSECONDS = 24 * 3600


def jlencode(iterable):
    if isinstance(iterable, (dict, str, unicode)):
        iterable = [iterable]
    return u'\n'.join(jsonencode(o) for o in iterable)


def jldecode(lineiterable):
    for line in lineiterable:
        yield loads(line)


def jsonencode(o):
    return dumps(o, default=jsondefault)


def jsondefault(o):
    if isinstance(o, datetime):
        delta = o - EPOCH
        u = delta.microseconds
        s = delta.seconds
        d = delta.days
        millis = (u + (s + d * ADAYINSECONDS) * 1e6) / 1000
        return int(millis)
    else:
        return str(o)




import os
import unittest
import random
from hubstorage import HubstorageClient


class HSTestCase(unittest.TestCase):

    projectid = str(random.randint(2222000, 2223000))
    spidername = 'hs-test-spider'
    endpoint = os.getenv('HS_ENDPOINT', 'http://storage.vm.scrapinghub.com')
    auth = os.getenv('HS_AUTH', 'f' * 32)
    frontier = 'test'
    slot = 'site.com'
    testbotgroup = 'python-hubstorage-test'

    @classmethod
    def setUpClass(cls):
        cls.hsclient = HubstorageClient(auth=cls.auth, endpoint=cls.endpoint)
        cls.project = cls.hsclient.get_project(cls.projectid)
        cls.spiderid = str(cls.project.ids.spider(cls.spidername, create=1))
        cls._set_testbotgroup()

    def setUp(self):
        self._set_testbotgroup()
        self._remove_all_jobs()

    def tearDown(self):
        self._remove_all_jobs()

    @classmethod
    def tearDownClass(cls):
        cls.hsclient.close()
        cls._remove_all_jobs()
        cls._unset_testbotgroup()

    @classmethod
    def _remove_all_jobs(cls):
        project = cls.project
        for k in project.settings.keys():
            if k != 'botgroups':
                del project.settings[k]
        project.settings.save()

        # Cleanup JobQ
        jobq = project.jobq
        for queuename in ('pending', 'running', 'finished'):
            info = {'summary': [None]}  # poor-guy do...while
            while info['summary']:
                info = jobq.summary(queuename)
                for summary in info['summary']:
                    cls._remove_job(summary['key'])

    @classmethod
    def _remove_job(cls, jobkey):
        cls.project.jobq.delete(jobkey)
        cls._delete_job(jobkey)

    @classmethod
    def _delete_job(cls, jobkey):
        assert jobkey.startswith(cls.projectid), jobkey
        cls.project.jobs.apidelete(jobkey.partition('/')[2])

    @classmethod
    def _set_testbotgroup(cls):
        cls.project.settings.apipost(jl={'botgroups': [cls.testbotgroup]})
        cls.project.settings.expire()

    @classmethod
    def _unset_testbotgroup(cls):
        cls.project.settings.apidelete('botgroups')
        cls.project.settings.expire()


class NopTest(HSTestCase):

    def test_nop(self):
        pass  # hooray!


if __name__ == '__main__':
    unittest.main()



"""
Test Client
"""
from hstestcase import HSTestCase
from hubstorage.utils import millitime, apipoll

class ClientTest(HSTestCase):

    def test_push_job(self):
        c = self.hsclient
        c.push_job(self.projectid, self.spidername,
                   priority=self.project.jobq.PRIO_LOW,
                   foo='baz')
        job = c.start_job(projectid=self.projectid)
        m = job.metadata
        self.assertEqual(m.get('state'), u'running', c.auth)
        self.assertEqual(m.get('foo'), u'baz')
        self.project.jobq.delete(job)

        # job auth token is valid only while job is running
        m = c.get_job(job.key).metadata
        self.assertEqual(m.get('state'), u'deleted')
        self.assertEqual(m.get('foo'), u'baz')

    def test_botgroup(self):
        self.project.settings.update(botgroups=['foo'], created=millitime())
        self.project.settings.save()
        c = self.hsclient
        q1 = c.push_job(self.project.projectid, self.spidername)
        j1 = c.start_job()
        self.assertEqual(j1, None, 'got %s, pushed job was %s' % (j1, q1))
        j2 = c.start_job(botgroup='bar')
        self.assertEqual(j2, None, 'got %s, pushed job was %s' % (j2, q1))
        j3 = apipoll(self.hsclient.start_job, botgroup='foo')
        self.assertEqual(j3.key, q1.key)

    def test_start_job(self):
        # Pending queue is empty
        job = self.hsclient.start_job(botgroup=self.testbotgroup)
        self.assertEqual(job, None)
        # Push a new job into pending queue
        j0 = self.hsclient.push_job(self.projectid, self.spidername)
        # Assert it is pending
        self.assertEqual(j0.metadata.get('state'), 'pending')
        # Poll for the pending job
        j1 = apipoll(self.hsclient.start_job, botgroup=self.testbotgroup)
        # Assert started job does not need an extra request to fetch its metadata
        self.assertTrue(j1.metadata._cached is not None)
        # Assert started job is in running queue
        self.assertEqual(j1.metadata.get('state'), 'running')
        # Started job metadata must match metadata retrieved directly from /jobs/
        j2 = self.hsclient.get_job(j1.key)
        self.assertEqual(dict(j1.metadata), dict(j2.metadata))

    def test_jobsummaries(self):
        hsc = self.hsclient
        # add at least one running or pending job to ensure summary is returned
        hsc.push_job(self.projectid, self.spidername, state='running')

        def _get_summary():
            jss = hsc.projects.jobsummaries()
            mjss = dict((str(js['project']), js) for js in jss)
            return mjss.get(self.projectid)
        summary = apipoll(_get_summary)
        self.assertIsNotNone(summary)

    def test_timestamp(self):
        ts1 = self.hsclient.server_timestamp()
        ts2 = self.hsclient.server_timestamp()
        self.assertGreater(ts1, 0)
        self.assertLessEqual(ts1, ts2)



import json

from .resourcetype import ResourceType
from .utils import urlpathjoin


class Frontier(ResourceType):

    resource_type = 'hcf'

    batch_size = 5000
    batch_qsize = 6000  # defaults to twice batch_size if None
    batch_start = 0
    batch_interval = 60.0
    batch_append = False
    batch_content_encoding = 'identity'

    def __init__(self, *a, **kw):
        self._writers = {}  # dict of writers indexed by (frontier, slot)
        self.newcount = 0
        super(Frontier, self).__init__(*a, **kw)

    def _get_writer(self, frontier, slot):
        key = (frontier, slot)
        writer = self._writers.get(key)
        if not writer:
            writer = self.client.batchuploader.create_writer(
                url=urlpathjoin(self.url, frontier, 's', slot),
                auth=self.auth,
                size=self.batch_size,
                start=self.batch_start,
                interval=self.batch_interval,
                qsize=self.batch_qsize,
                content_encoding=self.batch_content_encoding,
                callback=self._writer_callback
            )
            self._writers[key] = writer
        return writer

    def _writer_callback(self, response):
        self.newcount += json.loads(response.content)["newcount"]

    def close(self, block=True):
        for writer in self._writers.values():
            writer.close(block=block)

    def flush(self):
        for writer in self._writers.values():
            writer.flush()

    def add(self, frontier, slot, fps):
        writer = self._get_writer(frontier, slot)
        for fp in fps:
            writer.write(fp)

    def read(self, frontier, slot, mincount=None):
        params = {}
        if mincount is not None:
            params['mincount'] = mincount
        return self.apiget((frontier, 's', slot, 'q'), params=params)

    def delete(self, frontier, slot, ids):
        self.apipost((frontier, 's', slot, 'q/deleted'), jl=ids)

    def delete_slot(self, frontier, slot):
        self.apidelete((frontier, 's', slot))



"""
High level Hubstorage client
"""
from httplib import BadStatusLine
import logging
import pkgutil
from requests import session, adapters, HTTPError, ConnectionError, Timeout
from retrying import Retrying
from .utils import xauth, urlpathjoin
from .project import Project
from .job import Job
from .jobq import JobQ
from .batchuploader import BatchUploader
from .resourcetype import ResourceType

__all__ = ["HubstorageClient"]
__version__ = pkgutil.get_data('hubstorage', 'VERSION').strip()


logger = logging.getLogger('HubstorageClient')

_HTTP_ERROR_CODES_TO_RETRY = (408, 429, 503, 504)


def _hc_retry_on_exception(err):
    """Callback used by the client to restrict the retry to acceptable errors"""
    if (isinstance(err, HTTPError) and err.response.status_code in _HTTP_ERROR_CODES_TO_RETRY):
        logger.warning("Server failed with %d status code, retrying (maybe)" % (err.response.status_code,))
        return True

    # TODO: python3 compatibility: BadStatusLine error are wrapped differently
    if (isinstance(err, ConnectionError) and err.args[0] == 'Connection aborted.' and
            isinstance(err.args[1], BadStatusLine) and err.args[1][0] == repr('')):
        logger.warning("Protocol failed with BadStatusLine, retrying (maybe)")
        return True

    if isinstance(err, Timeout):
        logger.warning("Server connection timeout, retrying (maybe)")
        return True

    return False

class HubstorageClient(object):

    DEFAULT_ENDPOINT = 'http://storage.scrapinghub.com/'
    USERAGENT = 'python-hubstorage/{0}'.format(__version__)

    DEFAULT_CONNECTION_TIMEOUT_S = 60.0
    RETRY_DEFAUT_MAX_RETRY_TIME_S = 60.0

    RETRY_DEFAULT_MAX_RETRIES = 3
    RETRY_DEFAULT_JITTER_MS = 500
    RETRY_DEFAULT_EXPONENTIAL_BACKOFF_MS = 500

    def __init__(self, auth=None, endpoint=None, connection_timeout=None, max_retries=None, max_retry_time=None):
        """
        Note:
            max_retries and max_retry_time change how the client attempt to retry failing requests that are
            idempotent (safe to execute multiple time).

            HubstorageClient(max_retries=3) will retry requests 3 times, no matter the time it takes.
            Use max_retry_time if you want to bound the time spent in retrying.

            By default, requests are retried at most 3 times, during 60 seconds.

        Args:
            auth (str): The client authentication token
            endpoint (str): The API root address
            connection_timeout (int): The connection timeout for a _single request_
            max_retries (int): The number of time idempotent requests may be retried
            max_retry_time (int): The time, in seconds, during which the client can retry a request
        """
        self.auth = xauth(auth)
        self.endpoint = endpoint or self.DEFAULT_ENDPOINT
        self.connection_timeout = connection_timeout or self.DEFAULT_CONNECTION_TIMEOUT_S
        self.session = self._create_session()
        self.retrier = self._create_retrier(max_retries, max_retry_time)
        self.jobq = JobQ(self, None)
        self.projects = Projects(self, None)
        self.root = ResourceType(self, None)
        self._batchuploader = None

    def request(self, is_idempotent=False, **kwargs):
        """
        Execute an HTTP request with the current client session.

        Use the retry policy configured in the client when is_idempotent is True
        """
        kwargs.setdefault('timeout', self.connection_timeout)

        def invoke_request():
            r = self.session.request(**kwargs)

            if not r.ok:
                logger.debug('%s: %s', r, r.content)
            r.raise_for_status()
            return r

        if is_idempotent:
            return self.retrier.call(invoke_request)
        else:
            return invoke_request()

    def _create_retrier(self, max_retries, max_retry_time):
        """
        Create the Retrier object used to process idempotent client requests.

        If only max_retries is set, the default max_retry_time is ignored.

        Args:
            max_retries (int): the number of retries to be attempted
            max_retry_time (int): the number of time, in seconds, to retry for.
        Returns:
            A Retrying instance, that implements a call(func) method.
        """

        # Client sets max_retries only
        if max_retries is not None and max_retry_time is None:
            stop_max_delay = None
            stop_max_attempt_number = max_retries + 1
            wait_exponential_multiplier = self.RETRY_DEFAULT_EXPONENTIAL_BACKOFF_MS
        else:
            stop_max_delay = (max_retry_time or self.RETRY_DEFAUT_MAX_RETRY_TIME_S) * 1000.0
            stop_max_attempt_number = (max_retries or self.RETRY_DEFAULT_MAX_RETRIES) + 1

            # Compute the backoff to allow for max_retries queries during the allowed delay
            # Solves the following formula (assumes requests are immediate):
            # max_retry_time = sum(exp_multiplier * 2 ** i) for i from 1 to max_retries + 1
            wait_exponential_multiplier = stop_max_delay / ((2 ** (stop_max_attempt_number + 1)) - 2)

        return Retrying(stop_max_attempt_number=stop_max_attempt_number,
                        stop_max_delay=stop_max_delay,
                        retry_on_exception=_hc_retry_on_exception,
                        wait_exponential_multiplier=wait_exponential_multiplier,
                        wait_jitter_max=self.RETRY_DEFAULT_JITTER_MS)

    def _create_session(self):
        s = session()
        s.headers.update({'User-Agent': self.USERAGENT})
        return s

    @property
    def batchuploader(self):
        if self._batchuploader is None:
            self._batchuploader = BatchUploader(self)
        return self._batchuploader

    def get_job(self, *args, **kwargs):
        return Job(self, *args, **kwargs)

    def push_job(self, projectid, spidername, auth=None, **jobparams):
        project = self.projects.get(projectid, auth=auth)
        return project.push_job(spidername, **jobparams)

    def start_job(self, projectid=None, auth=None, **startparams):
        """Start a new job

        It may take up to a second for a previously added job to be available
        here if no project id is specified.

        """
        if projectid:
            jobq = self.projects.get(projectid, auth=auth).jobq
        else:
            jobq = self.jobq
        jobdata = jobq.start(**startparams)
        if jobdata:
            jobkey = jobdata.pop('key')
            jobauth = (jobkey, jobdata['auth'])
            return self.get_job(jobkey, jobauth=jobauth, metadata=jobdata)

    def get_project(self, *args, **kwargs):
        return self.projects.get(*args, **kwargs)

    def server_timestamp(self):
        tsurl = urlpathjoin(self.endpoint, 'system/ts')
        return self.session.get(tsurl).json()

    def close(self, timeout=None):
        if self._batchuploader is not None:
            self.batchuploader.close(timeout)


class Projects(ResourceType):

    resource_type = 'projects'

    def get(self, *args, **kwargs):
        return Project(self.client, *args, **kwargs)

    def jobsummaries(self, auth=None, **params):
        auth = xauth(auth) or self.auth
        return next(self.apiget('jobsummaries', params=params, auth=auth))



"""
Test Activty
"""
from hstestcase import HSTestCase


class ActivityTest(HSTestCase):

    def test_post_and_reverse_get(self):
        # make some sample data
        orig_data = [{u'foo': 42, u'counter': i} for i in xrange(20)]
        data1 = orig_data[:10]
        data2 = orig_data[10:]

        # put ordered data in 2 separate posts
        self.project.activity.post(data1)
        self.project.activity.post(data2)

        # read them back in reverse chronological order
        result = list(self.project.activity.list(count=20))
        self.assertEqual(len(result), 20)
        self.assertEqual(orig_data[::-1], result)

    def test_filters(self):
        self.project.activity.post({'c': i} for i in xrange(10))
        r = list(self.project.activity.list(filter='["c", ">", [5]]', count=2))
        self.assertEqual(r, [{'c': 9}, {'c': 8}])

    def test_timestamp(self):
        self.project.activity.add({'foo': 'bar'}, baz='qux')
        entry = self.project.activity.list(count=1, meta='_ts').next()
        self.assertTrue(entry.pop('_ts', None))
        self.assertEqual(entry, {'foo': 'bar', 'baz': 'qux'})



"""
Test job metadata

System tests for operations on stored job metadata
"""
from hstestcase import HSTestCase


class JobsMetadataTest(HSTestCase):

    def _assertMetadata(self, meta1, meta2):
        def _clean(m):
            return dict((k, v) for k, v in m.items() if k != 'updated_time')

        meta1 = _clean(meta1)
        meta2 = _clean(meta2)
        self.assertEqual(meta1, meta2)

    def test_basic(self):
        job = self.project.push_job(self.spidername)
        self.assertTrue('auth' not in job.metadata)
        self.assertTrue('state' in job.metadata)
        self.assertEqual(job.metadata['spider'], self.spidername)

        # set some metadata and forget it
        job.metadata['foo'] = 'bar'
        self.assertEqual(job.metadata['foo'], 'bar')
        job.metadata.expire()
        self.assertTrue('foo' not in job.metadata)

        # set it again and persist it
        job.metadata['foo'] = 'bar'
        self.assertEqual(job.metadata['foo'], 'bar')
        job.metadata.save()
        self.assertEqual(job.metadata['foo'], 'bar')
        job.metadata.expire()
        self.assertEqual(job.metadata['foo'], 'bar')

        # refetch the job and compare its metadata
        job2 = self.hsclient.get_job(job.key)
        self._assertMetadata(job2.metadata, job.metadata)

        # delete foo but do not persist it
        del job.metadata['foo']
        self.assertTrue('foo' not in job.metadata)
        job.metadata.expire()
        self.assertEqual(job.metadata.get('foo'), 'bar')
        # persist it to be sure it is not removed
        job.metadata.save()
        self.assertEqual(job.metadata.get('foo'), 'bar')
        # and finally delete again and persist it
        del job.metadata['foo']
        self.assertTrue('foo' not in job.metadata)
        job.metadata.save()
        self.assertTrue('foo' not in job.metadata)
        job.metadata.expire()
        self.assertTrue('foo' not in job.metadata)

        job2 = self.hsclient.get_job(job.key)
        self._assertMetadata(job.metadata, job2.metadata)

    def test_updating(self):
        job = self.project.push_job(self.spidername)
        self.assertIsNone(job.metadata.get('foo'))
        job.update_metadata({'foo': 'bar'})
        # metadata attr should change
        self.assertEqual(job.metadata.get('foo'), 'bar')
        # as well as actual metadata
        job = self.project.get_job(job.key)
        self.assertEqual(job.metadata.get('foo'), 'bar')
        job.update_metadata({'foo': None})
        self.assertFalse(job.metadata.get('foo', False))

        # there are ignored fields like: auth, _key, state
        state = job.metadata['state']
        job.update_metadata({'state': 'running'})
        self.assertEqual(job.metadata['state'], state)

    def test_representation(self):
        job = self.project.push_job(self.spidername)
        meta = job.metadata
        self.assertNotEqual(str(meta), repr(meta))
        self.assertEqual(meta, eval(str(meta)))
        self.assertTrue(meta.__class__.__name__ in repr(meta))
        self.assertFalse(meta.__class__.__name__ in str(meta))

    def test_jobauth(self):
        job = self.project.push_job(self.spidername)
        self.assertIsNone(job.jobauth)
        self.assertEqual(job.auth, self.project.auth)
        self.assertEqual(job.items.auth, self.project.auth)

        samejob = self.hsclient.get_job(job.key)
        self.assertIsNone(samejob.auth)
        self.assertIsNone(samejob.jobauth)
        self.assertEqual(samejob.items.auth, self.project.auth)

    def test_authtoken(self):
        pendingjob = self.project.push_job(self.spidername)
        runningjob = self.project.start_job()
        self.assertEqual(pendingjob.key, runningjob.key)
        self.assertTrue(runningjob.jobauth)
        self.assertEqual(runningjob.jobauth, runningjob.auth)
        self.assertEqual(runningjob.auth[0], runningjob.key)
        self.assertTrue(runningjob.auth[1])



"""
Test Retry Policy
"""
from httplib import BadStatusLine
from requests import HTTPError, ConnectionError
from hstestcase import HSTestCase
from hubstorage import HubstorageClient
import responses
import json
import re

GET = responses.GET
POST = responses.POST
DELETE = responses.DELETE


class RetryTest(HSTestCase):
    def test_delete_on_hubstorage_api_does_not_404(self):
        # NOTE: The current Hubstorage API does not raise 404 errors on deleting resources that do not exist,
        #       Thus the retry policy does not catch 404 errors when retrying deletes (simplify implementation A LOT).
        #       This test checks that this assumption holds.

        client = HubstorageClient(auth=self.auth, endpoint=self.endpoint, max_retries=0)
        project = client.get_project(projectid=self.projectid)

        # Check frontier delete
        project.frontier.delete_slot('frontier_non_existing', 'slot_non_existing')

        # Check metadata delete
        job = client.push_job(self.projectid, self.spidername)
        job.metadata['foo'] = 'bar'  # Add then delete key, this will trigger an api delete for item foo
        del job.metadata['foo']
        job.metadata.save()

        # Check collections delete
        store = project.collections.new_store('foo')
        store.set({'_key': 'foo'})
        store.delete('bar')

        self.assertTrue(True, "No error have been triggered by calling a delete on resources that do not exist")

    @responses.activate
    def test_retrier_does_not_catch_unwanted_exception(self):
        # Prepare
        client = HubstorageClient(auth=self.auth, endpoint=self.endpoint, max_retries=2)
        job_metadata = {'project': self.projectid, 'spider': self.spidername, 'state': 'pending'}
        callback, attempts_count = self.make_request_callback(3, job_metadata, http_error_status=403)

        self.mock_api(callback=callback)

        # Act
        job, metadata, err = None, None, None
        try:
            job = client.get_job('%s/%s/%s' % (self.projectid, self.spiderid, 42))
            metadata = dict(job.metadata)
        except HTTPError as e:
            err = e

        # Assert
        self.assertIsNone(metadata)
        self.assertIsNotNone(err)
        self.assertEqual(err.response.status_code, 403)
        self.assertEqual(attempts_count[0], 1)

    @responses.activate
    def test_retrier_catches_badstatusline_and_429(self):
        # Prepare
        client = HubstorageClient(auth=self.auth, endpoint=self.endpoint, max_retries=3)
        job_metadata = {'project': self.projectid, 'spider': self.spidername, 'state': 'pending'}

        attempts_count = [0]  # use a list for nonlocal mutability used in request_callback

        def request_callback(request):
            attempts_count[0] += 1

            if attempts_count[0] <= 2:
                raise ConnectionError("Connection aborted.", BadStatusLine("''"))
            if attempts_count[0] == 3:
                return (429, {}, {})
            else:
                resp_body = dict(job_metadata)
                return (200, {}, json.dumps(resp_body))

        self.mock_api(callback=request_callback)

        # Act
        job = client.get_job('%s/%s/%s' % (self.projectid, self.spiderid, 42))

        # Assert
        self.assertEqual(dict(job_metadata), dict(job.metadata))
        self.assertEqual(attempts_count[0], 4)

    @responses.activate
    def test_api_delete_can_be_set_to_non_idempotent(self):
        # Prepare
        client = HubstorageClient(auth=self.auth, endpoint=self.endpoint, max_retries=3)
        job_metadata = {'project': self.projectid, 'spider': self.spidername, 'state': 'pending'}
        callback_delete, attempts_count_delete = self.make_request_callback(2, job_metadata)

        self.mock_api(method=DELETE, callback=callback_delete)

        # Act
        job = client.get_job('%s/%s/%s' % (self.projectid, self.spiderid, 42))

        err = None
        try:
            job.metadata.apidelete('/my/non/idempotent/delete/', is_idempotent=False)
        except HTTPError as e:
            err = e

        # Assert
        self.assertEqual(attempts_count_delete[0], 1)
        self.assertIsNotNone(err)

    @responses.activate
    def test_collection_store_and_delete_are_retried(self):
        # Prepare
        client = HubstorageClient(auth=self.auth, endpoint=self.endpoint, max_retries=3)

        callback_post, attempts_count_post = self.make_request_callback(2, [])
        callback_delete, attempts_count_delete = self.make_request_callback(2, [])

        self.mock_api(method=POST, callback=callback_delete, url_match='/.*/deleted')
        self.mock_api(method=POST, callback=callback_post)  # /!\ default regexp matches all paths, has to be added last

        # Act
        project = client.get_project(self.projectid)
        store = project.collections.new_store('foo')
        store.set({'_key': 'bar', 'content': 'value'})
        store.delete('baz')

        # Assert
        self.assertEqual(attempts_count_post[0], 3)
        self.assertEqual(attempts_count_delete[0], 3)

    @responses.activate
    def test_delete_requests_are_retried(self):
        # Prepare
        client = HubstorageClient(auth=self.auth, endpoint=self.endpoint, max_retries=3)
        job_metadata = {'project': self.projectid, 'spider': self.spidername, 'state': 'pending'}
        callback_getpost, attempts_count_getpost = self.make_request_callback(0, job_metadata)
        callback_delete, attempts_count_delete = self.make_request_callback(2, job_metadata)

        self.mock_api(method=GET, callback=callback_getpost)
        self.mock_api(method=POST, callback=callback_getpost)
        self.mock_api(method=DELETE, callback=callback_delete)

        # Act
        job = client.get_job('%s/%s/%s' % (self.projectid, self.spiderid, 42))
        job.metadata['foo'] = 'bar'
        del job.metadata['foo']
        job.metadata.save()

        # Assert
        self.assertEqual(attempts_count_delete[0], 3)

    @responses.activate
    def test_metadata_save_does_retry(self):
        # Prepare
        client = HubstorageClient(auth=self.auth, endpoint=self.endpoint, max_retries=3)
        job_metadata = {'project': self.projectid, 'spider': self.spidername, 'state': 'pending'}
        callback_get, attempts_count_get = self.make_request_callback(0, job_metadata)
        callback_post, attempts_count_post = self.make_request_callback(2, job_metadata)

        self.mock_api(method=GET, callback=callback_get)
        self.mock_api(method=POST, callback=callback_post)

        # Act
        job = client.get_job('%s/%s/%s' % (self.projectid, self.spiderid, 42))
        job.metadata['foo'] = 'bar'
        job.metadata.save()

        # Assert
        self.assertEqual(attempts_count_post[0], 3)

    @responses.activate
    def test_push_job_does_not_retry(self):
        # Prepare
        client = HubstorageClient(auth=self.auth, endpoint=self.endpoint, max_retries=3)
        callback, attempts_count = self.make_request_callback(2, {'key': '1/2/3'})

        self.mock_api(POST, callback=callback)

        # Act
        job, err = None, None
        try:
            job = client.push_job(self.projectid, self.spidername)
        except HTTPError as e:
            err = e

        # Assert
        self.assertIsNone(job)
        self.assertIsNotNone(err)
        self.assertEqual(err.response.status_code, 504)
        self.assertEqual(attempts_count[0], 1)

    @responses.activate
    def test_get_job_does_retry(self):
        # Prepare
        client = HubstorageClient(auth=self.auth, endpoint=self.endpoint, max_retries=3)
        job_metadata = {'project': self.projectid, 'spider': self.spidername, 'state': 'pending'}
        callback, attempts_count = self.make_request_callback(2, job_metadata)

        self.mock_api(callback=callback)

        # Act
        job = client.get_job('%s/%s/%s' % (self.projectid, self.spiderid, 42))

        # Assert
        self.assertEqual(dict(job_metadata), dict(job.metadata))
        self.assertEqual(attempts_count[0], 3)

    @responses.activate
    def test_get_job_does_fails_if_no_retries(self):
        # Prepare
        client = HubstorageClient(auth=self.auth, endpoint=self.endpoint, max_retries=0)
        job_metadata = {'project': self.projectid, 'spider': self.spidername, 'state': 'pending'}
        callback, attempts_count = self.make_request_callback(2, job_metadata)

        self.mock_api(callback=callback)

        # Act
        job, metadata, err = None, None, None
        try:
            job = client.get_job('%s/%s/%s' % (self.projectid, self.spiderid, 42))
            metadata = dict(job.metadata)
        except HTTPError as e:
            err = e

        # Assert
        self.assertIsNone(metadata)
        self.assertIsNotNone(err)
        self.assertEqual(err.response.status_code, 504)
        self.assertEqual(attempts_count[0], 1)

    @responses.activate
    def test_get_job_does_fails_on_too_many_retries(self):
        # Prepare
        client = HubstorageClient(auth=self.auth, endpoint=self.endpoint, max_retries=2)
        job_metadata = {'project': self.projectid, 'spider': self.spidername, 'state': 'pending'}
        callback, attempts_count = self.make_request_callback(3, job_metadata)

        self.mock_api(callback=callback)

        # Act
        job, metadata, err = None, None, None
        try:
            job = client.get_job('%s/%s/%s' % (self.projectid, self.spiderid, 42))
            metadata = dict(job.metadata)
        except HTTPError as e:
            err = e

        # Assert
        self.assertIsNone(metadata)
        self.assertIsNotNone(err)
        self.assertEqual(err.response.status_code, 504)
        self.assertEqual(attempts_count[0], 3)

    def mock_api(self, method=GET, callback=None, url_match='/.*'):
        """
        Mock an API URL using the responses library.

        Args:
            method (Optional[str]): The HTTP method to mock. Defaults to responses.GET
            callback (function(request) -> response):
            url_match (Optional[str]): The API URL regexp. Defaults to '/.*'.
        """
        responses.add_callback(
            method, re.compile(self.endpoint + url_match),
            callback=callback,
            content_type='application/json',
        )

    def make_request_callback(self, timeout_count, body_on_success, http_error_status=504):
        """
        Make a request callback that timeout a couple of time before returning body_on_success

        Returns:
            A tuple (request_callback, attempts), attempts is an array of size one that contains the number of time
            request_callback has been called.
        """
        attempts = [0]  # use a list for nonlocal mutability used in request_callback

        def request_callback(request):
            attempts[0] += 1

            if attempts[0] <= timeout_count:
                return (http_error_status, {}, "Timeout")
            else:
                resp_body = dict(body_on_success)
                return (200, {}, json.dumps(resp_body))

        return (request_callback, attempts)



import time
from Queue import Empty


def urlpathjoin(*parts):
    """Join multiple paths into a single url

    >>> urlpathjoin('http://storage.scrapinghub.com:8002/', 'jobs', '1/2/3')
    'http://storage.scrapinghub.com:8002/jobs/1/2/3'
    >>> urlpathjoin('http://storage.scrapinghub.com:8002', 'jobs', '1/2/3', None)
    'http://storage.scrapinghub.com:8002/jobs/1/2/3'
    >>> urlpathjoin(_, 'state')
    'http://storage.scrapinghub.com:8002/jobs/1/2/3/state'
    >>> urlpathjoin(_, None)
    'http://storage.scrapinghub.com:8002/jobs/1/2/3/state'
    >>> urlpathjoin(78)
    '78'
    >>> urlpathjoin('78')
    '78'
    >>> urlpathjoin('s', 78)
    's/78'
    >>> urlpathjoin('s', 78, 'foo')
    's/78/foo'
    >>> urlpathjoin('s/78/foo')
    's/78/foo'
    >>> urlpathjoin((111, 'jobs'), 33)
    '111/jobs/33'
    >>> urlpathjoin('http://localhost:8003/', ('jobs', '1111111'), '2/1')
    'http://localhost:8003/jobs/1111111/2/1'

    """
    url = None
    for p in parts:
        if p is None:
            continue
        elif isinstance(p, tuple):
            p = urlpathjoin(*p)
        elif isinstance(p, unicode):
            p = p.encode('utf8')
        elif not isinstance(p, str):
            p = str(p)

        url = p if url is None else '{0}/{1}'.format(url.rstrip('/'), p)

    return url


def xauth(auth):
    """Expand authentification token

    >>> xauth(None)
    >>> xauth(('user', 'pass'))
    ('user', 'pass')
    >>> xauth('user:pass')
    ('user', 'pass')
    >>> xauth('apikey')
    ('apikey', '')

    """
    if auth is None or isinstance(auth, tuple):
        return auth
    else:
        u, _, p = auth.partition(':')
        return u, p


def millitime(*a, **kw):
    """The difference, measured in milliseconds, between the current time
    and midnight, January 1, 1970 UTC.

    >>> e = millitime()
    >>> type(e)
    <type 'int'>
    """
    ts = time.time(*a, **kw)
    return int(ts * 1000)


class iterqueue(object):
    """Iterate a queue til a maximum number of messages are read or the queue is empty

    it exposes an attribute "count" with the number of messages read

    >>> from Queue import Queue
    >>> q = Queue()
    >>> for x in xrange(10):
    ...     q.put(x)
    >>> qiter = iterqueue(q)
    >>> list(qiter)
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    >>> qiter.count
    10

    >>> for x in xrange(10):
    ...     q.put(x)
    >>> qiter = iterqueue(q, maxcount=4)
    >>> list(qiter)
    [0, 1, 2, 3]
    >>> qiter.count
    4
    """

    def __init__(self, queue, maxcount=None):
        self.queue = queue
        self.maxcount = maxcount
        self.count = 0

    def __iter__(self):
        while (self.maxcount is None) or (self.count < self.maxcount):
            try:
                yield self.queue.get_nowait()
                self.count += 1
            except Empty:
                break


def apipoll(endpoint, *args, **kwargs):
    """Poll an api endpoint until there is a result that is not None

    poll_wait and max_poll can be specified in kwargs to set the polling
    interval and max wait time in seconds.
    """
    result = endpoint(*args, **kwargs)
    if result is not None:
        return result
    start = time.time()
    while True:
        poll_wait = kwargs.get('poll_wait', 1)
        max_poll = kwargs.get('max_poll', 60)
        time.sleep(poll_wait)
        result = endpoint(*args, **kwargs)
        if result is not None or (time.time() - start) > max_poll:
            return result




"""
Test Project
"""
import json
from random import randint, random
from requests.exceptions import HTTPError
from hubstorage import HubstorageClient
from hubstorage.utils import millitime
from hstestcase import HSTestCase
from testutil import failing_downloader


class ProjectTest(HSTestCase):

    def test_projectid(self):
        p1 = self.hsclient.get_project(int(self.projectid))
        p2 = self.hsclient.get_project(str(self.projectid))
        self.assertEqual(p1.projectid, p2.projectid)
        self.assertEqual(type(p1.projectid), str)
        self.assertEqual(type(p2.projectid), str)
        self.assertRaises(AssertionError, self.hsclient.get_project, '111/3')

    def test_get_job_from_key(self):
        job = self.project.push_job(self.spidername)
        parts = tuple(job.key.split('/'))
        self.assertEqual(len(parts), 3)
        self.assertEqual(parts[:2], (self.projectid, self.spiderid))
        samejob1 = self.hsclient.get_job(job.key)
        samejob2 = self.project.get_job(job.key)
        samejob3 = self.project.get_job(parts[1:])
        self.assertEqual(samejob1.key, job.key)
        self.assertEqual(samejob2.key, job.key)
        self.assertEqual(samejob3.key, job.key)

    def test_get_jobs(self):
        p = self.project
        j1 = p.push_job(self.spidername, testid=0)
        j2 = p.push_job(self.spidername, testid=1)
        j3 = p.push_job(self.spidername, testid=2)
        # global list must list at least one job
        self.assertTrue(list(p.get_jobs(count=1, state='pending')))
        # List all jobs for test spider
        r = list(p.get_jobs(spider=self.spidername, state='pending'))
        self.assertEqual([j.key for j in r], [j3.key, j2.key, j1.key])

    def test_get_jobs_with_legacy_filter(self):
        p = self.project
        j1 = p.push_job(self.spidername, state='finished',
                        close_reason='finished', tags=['t2'])
        j2 = p.push_job(self.spidername, state='finished',
                        close_reason='finished', tags=['t1'])
        j3 = p.push_job(self.spidername, state='pending')
        j4 = p.push_job(self.spidername, state='finished',
                        close_reason='failed', tags=['t1'])
        j5 = p.push_job(self.spidername + 'skip', state='finished',
                        close_reason='failed', tags=['t1'])

        filters = [['spider', '=', [self.spidername]],
                   ['state', '=', ['finished']],
                   ['close_reason', '=', ['finished']],
                   ['tags', 'haselement', ['t1']],
                   ['tags', 'hasnotelement', ['t2']]]
        jobs = p.get_jobs(filter=[json.dumps(x) for x in filters])
        assert [j.key for j in jobs] == [j2.key], jobs

    def test_push_job(self):
        job = self.project.push_job(self.spidername, state='running',
                                   priority=self.project.jobq.PRIO_HIGH,
                                   foo=u'bar')
        self.assertEqual(job.metadata.get('state'), u'running')
        self.assertEqual(job.metadata.get('foo'), u'bar')
        self.project.jobq.delete(job)
        job.metadata.expire()
        self.assertEqual(job.metadata.get('state'), u'deleted')
        self.assertEqual(job.metadata.get('foo'), u'bar')

    def test_auth(self):
        # client without global auth set
        hsc = HubstorageClient(endpoint=self.hsclient.endpoint)
        self.assertEqual(hsc.auth, None)

        # check no-auth access
        try:
            hsc.push_job(self.projectid, self.spidername)
        except HTTPError as exc:
            self.assertTrue(exc.response.status_code, 401)
        else:
            self.assertTrue(False, '401 not raised')

        try:
            hsc.get_project(self.projectid).push_job(self.spidername)
        except HTTPError as exc:
            self.assertTrue(exc.response.status_code, 401)
        else:
            self.assertTrue(False, '401 not raised')

        try:
            hsc.get_job((self.projectid, 1, 1)).items.list()
        except HTTPError as exc:
            self.assertTrue(exc.response.status_code, 401)
        else:
            self.assertTrue(False, '401 not raised')

        try:
            hsc.get_project(self.projectid).get_job((self.projectid, 1, 1)).items.list()
        except HTTPError as exc:
            self.assertTrue(exc.response.status_code, 401)
        else:
            self.assertTrue(False, '401 not raised')

        # create project with auth
        auth = self.hsclient.auth
        project = hsc.get_project(self.projectid, auth)
        self.assertEqual(project.auth, auth)
        job = project.push_job(self.spidername)
        samejob = project.get_job(job.key)
        self.assertEqual(samejob.key, job.key)

    def test_broad(self):
        project = self.hsclient.get_project(self.projectid)
        # populate project with at least one job
        job = project.push_job(self.spidername)
        self.assertEqual(job.metadata.get('state'), 'pending')
        job = project.start_job()
        self.assertEqual(job.metadata.get('state'), 'running')
        job.items.write({'title': 'bar'})
        job.logs.info('nice to meet you')
        job.samples.write([1, 2, 3])
        job.finished()

        # keep a jobid for get_job and unreference job
        jobid = job.key
        jobauth = job.auth
        del job

        self.assertTrue(list(project.items.list(self.spiderid, count=1)))
        self.assertTrue(list(project.logs.list(self.spiderid, count=1)))
        self.assertTrue(list(project.samples.list(self.spiderid, count=1)))

        job = project.client.get_job(jobid, jobauth=jobauth)
        job.purged()

    def test_settings(self):
        project = self.hsclient.get_project(self.projectid)
        settings = dict(project.settings)
        settings.pop('botgroups', None)  # ignore testsuite botgroups
        self.assertEqual(settings, {})
        project.settings['created'] = created = millitime()
        project.settings['botgroups'] = ['g1']
        project.settings.save()
        self.assertEqual(project.settings.liveget('created'), created)
        self.assertEqual(project.settings.liveget('botgroups'), ['g1'])
        project.settings.expire()
        self.assertEqual(dict(project.settings), {
            'created': created,
            'botgroups': ['g1'],
        })

    def test_requests(self):
        ts = millitime()
        job = self.project.push_job(self.spidername, state='running')
        # top parent
        r1 = job.requests.add(url='http://test.com/', status=200, method='GET',
                              rs=1337, duration=5, parent=None, ts=ts)
        # first child
        r2 = job.requests.add(url='http://test.com/2', status=400, method='POST',
                              rs=0, duration=1, parent=r1, ts=ts + 1)
        # another child with fingerprint set
        r3 = job.requests.add(url='http://test.com/3', status=400, method='PUT',
                              rs=0, duration=1, parent=r1, ts=ts + 2, fp='1234')

        job.close_writers()
        rr = job.requests.list()
        self.assertEqual(rr.next(),
                         {u'status': 200, u'rs': 1337,
                          u'url': u'http://test.com/', u'time': ts,
                          u'duration': 5, u'method': u'GET'})
        self.assertEqual(rr.next(),
                         {u'status': 400, u'parent': 0, u'rs': 0,
                          u'url': u'http://test.com/2', u'time': ts + 1,
                          u'duration': 1, u'method': u'POST'})
        self.assertEqual(rr.next(),
                         {u'status': 400, u'fp': u'1234', u'parent': 0,
                          u'rs': 0, u'url': u'http://test.com/3',
                          u'time': ts + 2, u'duration': 1,
                          u'method': u'PUT'})

        self.assertRaises(StopIteration, rr.next)

    def test_samples(self):
        # no samples stored
        j1 = self.project.push_job(self.spidername, state='running')
        self.assertEqual(list(j1.samples.list()), [])
        # simple fill
        ts = millitime()
        j1.samples.write([ts, 1, 2, 3])
        j1.samples.write([ts + 1, 5, 9, 4])
        j1.samples.flush()
        o = list(j1.samples.list())
        self.assertEqual(len(o), 2)
        self.assertEqual(o[0], [ts, 1, 2, 3])
        self.assertEqual(o[1], [ts + 1, 5, 9, 4])

        # random fill
        j2 = self.project.push_job(self.spidername, state='running')
        samples = []
        ts = millitime()
        count = int(j2.samples.batch_size * (random() + randint(1, 5)))
        for _ in xrange(count):
            ts += randint(1, 2**16)
            row = [ts] + list(randint(0, 2**16) for _ in xrange(randint(0, 100)))
            samples.append(row)
            j2.samples.write(row)
        j2.samples.flush()
        o = list(j2.samples.list())
        self.assertEqual(len(o), count)
        for r1, r2 in zip(samples, o):
            self.assertEqual(r1, r2)

    def test_jobsummary(self):
        js = self.project.jobsummary()
        self.assertEqual(js.get('project'), int(self.project.projectid), js)
        self.assertEqual(js.get('has_capacity'), True, js)
        self.assertTrue('pending' in js, js)
        self.assertTrue('running' in js, js)

    def test_bulkdata(self):
        j = self.project.push_job(self.spidername, state='running')
        for i in xrange(20):
            j.logs.info("log line %d" % i)
            j.items.write(dict(field1="item%d" % i))
            j.requests.add("http://test.com/%d" % i,
                200, 'GET', 10, None, 10, 120)
        for resourcename in ('logs', 'items', 'requests'):
            resource = getattr(j, resourcename)
            resource.flush()

            # downloading resource, with simulated failures
            with failing_downloader(resource):
                downloaded = list(resource.iter_values())
                self.assertEqual(len(downloaded), 20)



"""
Test Project
"""
import time
from collections import defaultdict
from hstestcase import HSTestCase
from hubstorage import ValueTooLarge


class BatchUploaderTest(HSTestCase):

    def _job_and_writer(self, **writerargs):
        self.project.push_job(self.spidername)
        job = self.project.start_job()
        bu = self.hsclient.batchuploader
        w = bu.create_writer(job.items.url, auth=self.auth, **writerargs)
        return job, w

    def test_writer_batchsize(self):
        job, w = self._job_and_writer(size=10)
        for x in xrange(111):
            w.write({'x': x})
        w.close()
        # this works only for small batches (previous size=10 and small data)
        # as internally HS may commit a single large request as many smaller
        # commits, each with different timestamps
        groups = defaultdict(int)
        for doc in job.items.list(meta=['_ts']):
            groups[doc['_ts']] += 1

        self.assertEqual(len(groups), 12)

    def test_writer_maxitemsize(self):
        job, w = self._job_and_writer()
        m = w.maxitemsize
        self.assertRaisesRegexp(
            ValueTooLarge,
            'Value exceeds max encoded size of 1048576 bytes:'
            ' \'{"b": "x+\\.\\.\\.\'',
            w.write, {'b': 'x' * m})
        self.assertRaisesRegexp(
            ValueTooLarge,
            'Value exceeds max encoded size of 1048576 bytes:'
            ' \'{"b+\\.\\.\\.\'',
            w.write, {'b'*m: 'x'})
        self.assertRaisesRegexp(
            ValueTooLarge,
            'Value exceeds max encoded size of 1048576 bytes:'
            ' \'{"b+\\.\\.\\.\'',
            w.write, {'b'*(m/2): 'x'*(m/2)})

    def test_writer_contentencoding(self):
        for ce in ('identity', 'gzip'):
            job, w = self._job_and_writer(content_encoding=ce)
            for x in xrange(111):
                w.write({'x': x})
            w.close()
            self.assertEqual(job.items.stats()['totals']['input_values'], 111)

    def test_writer_interval(self):
        job, w = self._job_and_writer(size=1000, interval=1)
        for x in xrange(111):
            w.write({'x': x})
            if x == 50:
                time.sleep(2)

        w.close()
        groups = defaultdict(int)
        for doc in job.items.list(meta=['_ts']):
            groups[doc['_ts']] += 1

        self.assertEqual(len(groups), 2)



from requests.exceptions import HTTPError
from .resourcetype import DownloadableResource
from .utils import urlpathjoin


class Collections(DownloadableResource):

    resource_type = 'collections'

    def get(self, _type, _name, _key=None, **params):
        try:
            r = self.apiget((_type, _name, _key), params=params)
            return r if _key is None else r.next()
        except HTTPError as exc:
            if exc.response.status_code == 404:
                raise KeyError(_key)
            elif exc.response.status_code == 400:
                raise ValueError(exc.response.text)
            else:
                raise

    def set(self, _type, _name, _values):
        try:
            return self.apipost((_type, _name), is_idempotent=True, jl=_values)
        except HTTPError as exc:
            if exc.response.status_code in (400, 413):
                raise ValueError(exc.response.text)
            else:
                raise

    def delete(self, _type, _name, _keys):
        return self.apipost((_type, _name, 'deleted'), is_idempotent=True, jl=_keys)

    def iter_json(self, _type, _name, requests_params=None, **apiparams):
        return DownloadableResource.iter_json(self, (_type, _name),
            requests_params=requests_params, **apiparams)

    def create_writer(self, coltype, colname, **writer_kwargs):
        kwargs = dict(writer_kwargs)
        kwargs.setdefault('content_encoding', 'gzip')
        kwargs.setdefault('auth', self.auth)
        url = urlpathjoin(self.url, coltype, colname)
        return self.client.batchuploader.create_writer(url,
            **kwargs)

    def new_collection(self, coltype, colname):
        return Collection(coltype, colname, self)

    def new_store(self, colname):
        return self.new_collection('s', colname)

    def new_cached_store(self, colname):
        return self.new_collection('cs', colname)

    def new_versioned_store(self, colname):
        return self.new_collection('vs', colname)

    def new_versioned_cached_store(self, colname):
        return self.new_collection('vcs', colname)

    def count(self, _type, _name, **params):
        return self._batch('GET', (_type, _name, 'count'), 'count', **params)

    def _batch(self, method, path, total_param, progress=None, **params):
        total = 0
        getparams = dict(params)
        try:
            while True:
                r = self.apirequest(path, method=method,
                    params=getparams).next()
                total += r[total_param]
                next = r.get('nextstart')
                if next is None:
                    break
                getparams['start'] = next
                if progress:
                    progress(total, next)
            return total
        except HTTPError as exc:
            if exc.response.status_code == 400:
                raise ValueError(exc.response.text)
            else:
                raise


class Collection(object):

    def __init__(self, coltype, colname, collections):
        self.coltype = coltype
        self.colname = colname
        self._collections = collections

    def create_writer(self, **kwargs):
        """Create a writer for async writing of bulk data

        kwargs are passed to batchuploader.create_writer, but auth and gzip
        content encoding are specified if not provided
        """
        return self._collections.create_writer(self.coltype, self.colname,
            **kwargs)

    def get(self, *args, **kwargs):
        return self._collections.get(self.coltype, self.colname, *args, **kwargs)

    def set(self, *args, **kwargs):
        return self._collections.set(self.coltype, self.colname, *args, **kwargs)

    def delete(self, *args, **kwargs):
        return self._collections.delete(self.coltype, self.colname, *args, **kwargs)

    def count(self, *args, **kwargs):
        return self._collections.count(self.coltype, self.colname, *args, **kwargs)

    def iter_json(self, requests_params=None, **apiparams):
        return self._collections.iter_json(self.coltype, self.colname,
            requests_params=requests_params, **apiparams)

    def iter_values(self, requests_params=None, **apiparams):
        return self._collections.iter_values(self.coltype, self.colname,
            requests_params=requests_params, **apiparams)



from os.path import join, dirname
from setuptools import setup, find_packages

# We can't import hubstorage.__version__ because it imports "requests" and it
# can not be available yet
__version__ = open(join(dirname(__file__), 'hubstorage/VERSION')).read().strip()


setup(name='hubstorage',
      version=__version__,
      license='BSD',
      description='Client interface for Scrapinghub HubStorage',
      author='Scrapinghub',
      author_email='info@scrapinghub.com',
      url='http://scrapinghub.com',
      platforms=['Any'],
      packages=find_packages(),
      package_data={'hubstorage': ['VERSION']},
      install_requires=['requests'],
      classifiers=['Development Status :: 4 - Beta',
                   'License :: OSI Approved :: BSD License',
                   'Operating System :: OS Independent',
                   'Programming Language :: Python'],
      )



from .resourcetype import ResourceType

# TODO: remove backwards compatible methods


class Activity(ResourceType):

    resource_type = 'activity'

    def list(self, **params):
        return self.apiget(params=params)
    get = list

    def post(self, _value, **params):
        return self.apipost(jl=_value, params=params)

    def add(self, *args, **kwargs):
        entry = dict(*args, **kwargs)
        return self.post(entry)



"""
Test JobQ
"""
import os, unittest
from hstestcase import HSTestCase
from hubstorage.jobq import DuplicateJobError
from hubstorage.utils import apipoll


EXCLUSIVE = os.environ.get('EXCLUSIVE_STORAGE')


class JobqTest(HSTestCase):

    def test_push(self):
        jobq = self.project.jobq
        qjob = jobq.push(self.spidername)
        self.assertTrue('key' in qjob, qjob)
        self.assertTrue('auth' in qjob, qjob)

        job = self.hsclient.get_job(qjob['key'])
        self.assertEqual(job.metadata.get('state'), u'pending')
        self.assertEqual(job.metadata.get('spider'), self.spidername)
        self.assertEqual(job.metadata.get('auth'), qjob['auth'])

        jobq.start(job)
        job.metadata.expire()
        self.assertEqual(job.metadata.get('state'), u'running')

        jobq.finish(job)
        job.metadata.expire()
        self.assertEqual(job.metadata.get('state'), u'finished')

        jobq.delete(job)
        job.metadata.expire()
        self.assertEqual(job.metadata.get('state'), u'deleted')

    def test_push_with_extras(self):
        qjob = self.project.jobq.push(self.spidername, foo='bar', baz='fuu')
        job = self.hsclient.get_job(qjob['key'])
        self.assertEqual(job.metadata.get('foo'), u'bar')
        self.assertEqual(job.metadata.get('baz'), u'fuu')

    def test_push_with_priority(self):
        jobq = self.project.jobq
        qjob = jobq.push(self.spidername, priority=jobq.PRIO_HIGHEST)
        self.assertTrue('key' in qjob, qjob)
        self.assertTrue('auth' in qjob, qjob)

    def test_push_with_state(self):
        qjob = self.project.jobq.push(self.spidername, state='running')
        self.assertTrue('key' in qjob, qjob)
        self.assertTrue('auth' in qjob, qjob)
        job = self.hsclient.get_job(qjob['key'])
        self.assertEqual(job.metadata.get('state'), u'running')

    def test_push_with_unique(self):
        jobq = self.project.jobq
        # no unique key
        jobq.push(self.spidername)
        jobq.push(self.spidername)
        jobq.push(self.spidername, unique=None)
        jobq.push(self.spidername, unique=None)

        # unique key
        q1 = jobq.push(self.spidername, unique='h1')
        jobq.push(self.spidername, unique='h2')
        self.assertRaises(DuplicateJobError, jobq.push, self.spidername, unique='h1')
        jobq.finish(q1)
        self.assertRaises(DuplicateJobError, jobq.push, self.spidername, unique='h2')
        jobq.push(self.spidername, unique='h1')

    def test_startjob(self):
        jobq = self.project.jobq
        qj = jobq.push(self.spidername)
        nj = jobq.start()
        self.assertTrue(nj.pop('pending_time', None), nj)
        self.assertTrue(nj.pop('running_time', None), nj)
        self.assertTrue(nj.pop('auth', None), nj)
        self.assertEqual(nj[u'key'], qj['key'])
        self.assertEqual(nj[u'spider'], self.spidername)
        self.assertEqual(nj[u'state'], u'running')
        self.assertEqual(nj[u'priority'], jobq.PRIO_NORMAL)

    def test_startjob_with_extras(self):
        jobq = self.project.jobq
        pushextras = {
            'string': 'foo',
            'integer': 1,
            'float': 3.2,
            'mixedarray': ['b', 1, None, True, False, {'k': 'c'}],
            'emptyarray': [],
            'mapping': {'alpha': 5, 'b': 'B', 'cama': []},
            'emptymapping': {},
            'true': True,
            'false': False,
            'nil': None,
        }
        qj = jobq.push(self.spidername, **pushextras)
        startextras = dict(('s_' + k, v) for k, v in pushextras.iteritems())
        nj = jobq.start(**startextras)
        self.assertEqual(qj['key'], nj['key'])
        for k, v in dict(pushextras, **startextras).iteritems():
            if type(v) is float:
                self.assertAlmostEqual(nj.get(k), v)
            else:
                self.assertEqual(nj.get(k), v)

    def test_startjob_order(self):
        jobq = self.project.jobq
        q1 = jobq.push(self.spidername)
        q2 = jobq.push(self.spidername)
        q3 = jobq.push(self.spidername)
        self.assertEqual(jobq.start()['key'], q1['key'])
        self.assertEqual(jobq.start()['key'], q2['key'])
        self.assertEqual(jobq.start()['key'], q3['key'])

    def test_summary(self):
        jobq = self.project.jobq
        # push at least one job per state
        jobq.push(self.spidername)
        jobq.push(self.spidername, state='running')
        jobq.push(self.spidername, state='finished')
        summaries = dict((s['name'], s) for s in jobq.summary())
        self.assertEqual(set(summaries), set(['pending', 'running', 'finished']))
        self.assertTrue(jobq.summary('pending'))
        self.assertTrue(jobq.summary('running'))
        self.assertTrue(jobq.summary('finished'))

    def test_summary_jobmeta(self):
        jobq = self.project.jobq
        jobq.push(self.spidername, foo='bar', caz='fuu')
        pendings = jobq.summary('pending', jobmeta='foo')['summary']
        p1 = pendings[0]
        self.assertEqual(p1.get('foo'), 'bar')
        self.assertFalse('caz' in p1)

        pendings = jobq.summary('pending', jobmeta=['foo', 'caz'])['summary']
        p1 = pendings[0]
        self.assertEqual(p1.get('foo'), 'bar')
        self.assertEqual(p1.get('caz'), 'fuu')

    def test_summary_countstart(self):
        # push more than 5 jobs into same queue
        N = 20
        jobq = self.project.jobq
        for state in ('pending', 'running', 'finished'):
            for idx in xrange(N):
                jobq.push(self.spidername, state=state, idx=idx)

            s1 = jobq.summary(state)
            self.assertEqual(s1['count'], N)
            self.assertEqual(len(s1['summary']), 5)

            s2 = jobq.summary(state, count=N)
            self.assertEqual(len(s2['summary']), N)

            s3 = jobq.summary(state, start=N - 6, count=3)
            self.assertEqual([o['key'] for o in s3['summary']],
                             [o['key'] for o in s2['summary'][-6:-3]])

    def test_summaries_and_state_changes(self):
        jobq = self.project.jobq
        j1 = jobq.push(self.spidername)
        j2 = jobq.push(self.spidername)
        j3 = jobq.push(self.spidername)
        j4 = jobq.push(self.spidername, state='running')
        # check queue summaries
        self._assert_queue('pending', [j3, j2, j1])
        self._assert_queue('running', [j4])
        self._assert_queue('finished', [])
        # change job states
        jobq.start(j1)
        jobq.finish(j2)
        jobq.finish(j4)
        # check summaries again
        self._assert_queue('pending', [j3])
        self._assert_queue('running', [j1])
        self._assert_queue('finished', [j4, j2])
        # delete all jobs and check for empty summaries
        jobq.delete(j1)
        jobq.delete(j2)
        jobq.delete(j3)
        jobq.delete(j4)
        self._assert_queue('pending', [])
        self._assert_queue('running', [])
        self._assert_queue('finished', [])

    def test_list_with_state(self):
        jobq = self.project.jobq
        j1 = jobq.push(self.spidername, state='finished')
        j2 = jobq.push(self.spidername, state='running')
        j3 = jobq.push(self.spidername, state='pending')
        j4 = jobq.push(self.spidername, state='finished')
        # Only finished jobs by default
        assert _keys(jobq.list()) == _keys([j4, j1])
        assert _keys(jobq.list(state='finished')) == _keys([j4, j1])
        assert _keys(jobq.list(state='running')) == _keys([j2])
        assert _keys(jobq.list(state=['running', 'pending'])) == _keys([j3, j2])

    def test_list_with_count(self):
        jobq = self.project.jobq
        j1 = jobq.push(self.spidername, state='finished')  # NOQA
        j2 = jobq.push(self.spidername, state='finished')  # NOQA
        j3 = jobq.push(self.spidername, state='finished')
        j4 = jobq.push(self.spidername, state='finished')
        # fetch only the 2 most recent jobs
        assert _keys(jobq.list(count=2)) == _keys([j4, j3])

    def test_list_with_stop(self):
        jobq = self.project.jobq
        j1 = jobq.push(self.spidername, state='finished')
        j2 = jobq.push(self.spidername, state='finished')
        j3 = jobq.push(self.spidername, state='finished')
        j4 = jobq.push(self.spidername, state='finished')
        # test "stop" parameter
        # we should stop before the 4th finished job
        assert _keys(jobq.list(stop=j1['key'])) == _keys([j4, j3, j2])

    def test_list_with_tags(self):
        jobq = self.project.jobq
        j1 = jobq.push(self.spidername, state='finished', tags=['t1'])
        j2 = jobq.push(self.spidername, state='finished', tags=['t2'])
        j3 = jobq.push(self.spidername, state='finished', tags=['t1', 't2'])
        j4 = jobq.push(self.spidername, state='finished')
        assert _keys(jobq.list(has_tag='t1')) == _keys([j3, j1])
        assert _keys(jobq.list(has_tag=['t2', 't1'])) == _keys([j3, j2, j1])
        assert _keys(jobq.list(has_tag='t2', lacks_tag='t1')) == _keys([j2])
        assert _keys(jobq.list(lacks_tag=['t1', 't2'])) == _keys([j4])

    # endts is not implemented
    @unittest.expectedFailure
    def test_list_with_startts_endts(self):
        jobq = self.project.jobq
        j1 = jobq.push(self.spidername, state='finished')  # NOQA
        j2 = jobq.push(self.spidername, state='finished')
        j3 = jobq.push(self.spidername, state='finished')
        j4 = jobq.push(self.spidername, state='finished')  # NOQA
        # test "startts/endts" parameters
        # endts is not inclusive
        # so we should get the 2 in the middle out of 4
        timestamps = [j['ts'] for j in jobq.list()]
        jobs = jobq.list(startts=timestamps[2], endts=timestamps[0])
        assert _keys(jobs) == _keys([j3, j2])

    def _assert_queue(self, qname, jobs):
        summary = self.project.jobq.summary(qname, spiderid=self.spiderid)
        self.assertEqual(summary['name'], qname)
        self.assertEqual(summary['count'], len(jobs))
        self.assertEqual(len(summary['summary']), len(jobs))
        # Most recent jobs first
        self.assertEqual([s['key'] for s in summary['summary']],
                         [j['key'] for j in jobs])

    def test_simple_botgroups(self):
        self.project.settings['botgroups'] = ['g1']
        self.project.settings.save()
        pq = self.project.jobq
        hq = self.hsclient.jobq
        q1 = pq.push(self.spidername)
        self.assertEqual(hq.start(botgroup='g3'), None)
        self.assertEqual(apipoll(hq.start, botgroup='g1')['key'], q1['key'])

    @unittest.skipUnless(EXCLUSIVE, "test requires exclusive"
        " (without any active bots) access to HS. Set EXCLUSIVE_STORAGE"
        " env. var to activate")
    def test_botgroups(self):
        self.project.settings['botgroups'] = ['g1', 'g2']
        self.project.settings.save()
        pq = self.project.jobq
        hq = self.hsclient.jobq
        q1 = pq.push(self.spidername)
        q2 = pq.push(self.spidername)
        q3 = pq.push(self.spidername)
        self.assertEqual(hq.start(), None)
        self.assertEqual(hq.start(botgroup='g3'), None)
        self.assertEqual(apipoll(hq.start, botgroup='g1')['key'], q1['key'])
        self.assertEqual(apipoll(hq.start, botgroup='g2')['key'], q2['key'])

        # cleanup project botgroups, q3 must be polled only by generic bots
        del self.project.settings['botgroups']
        self.project.settings.save()
        q4 = pq.push(self.spidername)
        self.assertEqual(hq.start(botgroup='g1'), None)
        self.assertEqual(hq.start(botgroup='g2'), None)
        self.assertEqual(hq.start(botgroup='g3'), None)
        self.assertEqual(hq.start()['key'], q3['key'])
        self.assertEqual(hq.start()['key'], q4['key'])

        self.project.settings['botgroups'] = ['python-hubstorage-test']
        self.project.settings.save()

    def test_spider_updates(self):
        jobq = self.project.jobq
        spiderkey = '%s/%s' % (self.projectid, self.spiderid)
        q1 = jobq.push(self.spidername)
        q2 = jobq.push(self.spidername, state='running')
        q3 = jobq.push(self.spidername, state='finished')
        q4 = jobq.push(self.spidername, state='deleted')
        r = dict((x['key'], x['prevstate']) for x in jobq.delete(spiderkey))
        self.assertEqual(r.get(q1['key']), 'pending', r)
        self.assertEqual(r.get(q2['key']), 'running', r)
        self.assertEqual(r.get(q3['key']), 'finished', r)
        self.assertTrue(q4['key'] not in r)

        # Empty result set
        self.assertFalse(list(jobq.delete(spiderkey)))

    def test_multiple_job_update(self):
        jobq = self.project.jobq
        q1 = jobq.push(self.spidername)
        q2 = jobq.push(self.spidername)
        q3 = jobq.push(self.spidername)
        ids = [q1, q2['key'], self.project.get_job(q3['key'])]
        self.assertTrue([x['prevstate'] for x in jobq.start(ids)],
                        ['pending', 'pending', 'pending'])
        self.assertTrue([x['prevstate'] for x in jobq.finish(ids)],
                        ['running', 'running', 'running'])
        self.assertTrue([x['prevstate'] for x in jobq.delete(ids)],
                        ['finished', 'finished', 'finished'])


def _keys(lst):
    return [x['key'] for x in lst]



"""
Test Frontier
"""
from hstestcase import HSTestCase


class FrontierTest(HSTestCase):

    def setUp(self):
        self._delete_slot()

    def tearDown(self):
        self._delete_slot()

    def _delete_slot(self):
        frontier = self.project.frontier
        frontier.delete_slot(self.frontier, self.slot)

    def _remove_all_ids(self):
        frontier = self.project.frontier
        ids = [batch['id'] for batch in frontier.read(self.frontier, self.slot)]
        frontier.delete(self.frontier, self.slot, ids)

    def _get_urls(self, batch):
        return [r[0] for r in batch['requests']]

    def test_add_read(self):
        frontier = self.project.frontier

        fps = [{'fp': '/'}]
        frontier.add(self.frontier, self.slot, fps)
        fps = [{'fp': '/index.html'}, {'fp': '/index2.html'}]
        frontier.add(self.frontier, self.slot, fps)
        frontier.flush()

        urls = [self._get_urls(batch) for batch
                in frontier.read(self.frontier, self.slot)]
        expected_urls = [[u'/', u'/index.html', u'/index2.html']]
        self.assertEqual(urls, expected_urls)

    def test_add_multiple_chunks(self):
        frontier = self.project.frontier
        old_count = frontier.newcount

        batch_size = 50
        fps1 = [{'fp': '/index_%s.html' % fp} for fp in range(0, batch_size)]
        frontier.add(self.frontier, self.slot, fps1)

        fps2 = [{'fp': '/index_%s.html' % fp} for fp in range(batch_size, batch_size * 2)]
        frontier.add(self.frontier, self.slot, fps2)

        fps3 = [{'fp': '/index_%s.html' % fp} for fp in range(batch_size * 2, batch_size * 3)]
        frontier.add(self.frontier, self.slot, fps3)
        frontier.flush()

        self.assertEqual(frontier.newcount, 150 + old_count)

        # insert repeated fingerprints
        fps4 = [{'fp': '/index_%s.html' % fp} for fp in range(0, batch_size)]
        frontier.add(self.frontier, self.slot, fps3)
        frontier.flush()

        # new count is the same
        self.assertEqual(frontier.newcount, 150 + old_count)

        # get first 100
        batches = list(frontier.read(self.frontier, self.slot, mincount=100))
        urls = [self._get_urls(batch) for batch in batches]
        expected_urls = [[fp['fp'] for fp in fps1 + fps2]]
        self.assertEqual(urls, expected_urls)

        # delete first 100
        ids = [batch['id'] for batch in batches]
        frontier.delete(self.frontier, self.slot, ids)

        # get remaining 50
        batches = list(frontier.read(self.frontier, self.slot))
        urls = [self._get_urls(batch) for batch in batches]
        expected_urls = [[fp['fp'] for fp in fps3]]
        self.assertEqual(urls, expected_urls)

    def test_add_big_chunk(self):
        frontier = self.project.frontier

        batch_size = 300
        fps1 = [{'fp': '/index_%s.html' % fp} for fp in range(0, batch_size)]
        frontier.add(self.frontier, self.slot, fps1)
        frontier.flush()

        # get first 100
        batches = list(frontier.read(self.frontier, self.slot, mincount=100))
        urls = [self._get_urls(batch) for batch in batches]
        expected_urls = [[fp['fp'] for fp in fps1[:100]]]
        self.assertEqual(urls, expected_urls)

        # delete first 100
        ids = [batch['id'] for batch in batches]
        frontier.delete(self.frontier, self.slot, ids)

        # get next 100
        batches = list(frontier.read(self.frontier, self.slot, mincount=100))
        urls = [self._get_urls(batch) for batch in batches]
        expected_urls = [[fp['fp'] for fp in fps1[100:200]]]
        self.assertEqual(urls, expected_urls)

        # delete next 100
        ids = [batch['id'] for batch in batches]
        frontier.delete(self.frontier, self.slot, ids)

        # get next 100
        batches = list(frontier.read(self.frontier, self.slot, mincount=100))
        urls = [self._get_urls(batch) for batch in batches]
        expected_urls = [[fp['fp'] for fp in fps1[200:300]]]
        self.assertEqual(urls, expected_urls)

    def test_add_extra_params(self):
        frontier = self.project.frontier

        qdata = {"a": 1, "b": 2, "c": 3}
        fps = [{'fp': '/', "qdata": qdata}]
        frontier.add(self.frontier, self.slot, fps)
        frontier.flush()

        expected_request = [[u'/', {u'a': 1, u'c': 3, u'b': 2}]]
        batches = list(frontier.read(self.frontier, self.slot))
        request = batches[0]['requests']
        self.assertEqual(request, expected_request)

